/*****************************************************************************
|*
|*  VERSION CONTROL:    @(#)function.c	1.2 7/3/98
|*
|*  IN PACKAGE:         AX/HX93X Firmware
|*		 
|*  COPYRIGHT:          Copyright (c) 1998 Tasking Software B.V.
|*
|*  DESCRIPTION:
|*
|*	Universal serial bus AX/HX93X embedded function firmware
|*
 ****************************************************************************/

#include "eval.h"
#include "usb.h"
#include "function.h"
#include "loopback.h"

#if !defined( _C251 ) && !defined( _CC51 )
#define _using( X )
#endif

/*
 *	Function Config Descriptor structure type
 */
struct Conf
{
	CONFIG_DESCRIPTOR	config;
	INTERFACE_DESCRIPTOR	interface;
#if ( EPPS >= 1 ) && ( EPP_OFF < 1 )
	ENDPOINT_DESCRIPTOR	ep1tx;
	ENDPOINT_DESCRIPTOR	ep1rx;
#endif
#if ( EPPS >= 2 ) && ( EPP_OFF < 2 )
	ENDPOINT_DESCRIPTOR	ep2tx;
	ENDPOINT_DESCRIPTOR	ep2rx;
#endif
#if ( EPPS >= 3 ) && ( EPP_OFF < 3 )
	ENDPOINT_DESCRIPTOR	ep3tx;
	ENDPOINT_DESCRIPTOR	ep3rx;
#endif
#if defined( _REG930AX_SFR ) && ( EPPS >= 4 ) && ( EPP_OFF < 4 )
	ENDPOINT_DESCRIPTOR	ep4tx;
	ENDPOINT_DESCRIPTOR	ep4rx;
#endif
#if defined( _REG930AX_SFR ) && ( EPPS >= 5 ) && ( EPP_OFF < 5 )
	ENDPOINT_DESCRIPTOR	ep5tx;
	ENDPOINT_DESCRIPTOR	ep5rx;
#endif
};

typedef	struct Conf	CONF_DESCRIPTOR;

/*
 *	Descriptors
 */
static DEVICE_DESCRIPTOR device = 
{	sizeof(DEVICE_DESCRIPTOR),		/* size of this descriptor in bytes */
        DEVICE_DESCR,				/* device descriptor type */
	SWAP16(USB_SPEC), 			/* version 1.00 compliant */
        0,					/* independent interfaces */
	0,					/* with own class information */
	0,					/* no class specific protocols */
	EP0_MAX_PACKET_SIZE,			/* packet sizew 8 byte max for EP0 */
	SWAP16(USB_VENDOR_ID), 			/* Vendor ID */
	SWAP16(USB_PRODUCT_ID),			/* product ID */
	SWAP16(USB_DEVICE_ID),			/* device ID */
	MANUFACTURER_IDX, 			/* index string descriptor manufacturer */
	PRODUCT_IDX,				/* index string descriptor product */
       	0,//SERIAL_IDX,				/* index string serial number, No serial number */
	1					/* one configuration */
};

/*
 *	Initialize Config 1 descriptor 
 */
static CONF_DESCRIPTOR config1 =
{	{ 	sizeof(CONFIG_DESCRIPTOR),	/* size of this descriptor in bytes */
		CONFIG_DESCR,			/* configuration descriptor type */
		SWAP16(sizeof(CONF_DESCRIPTOR)),/* combined descriptor size */
		1,				/* one interface supported by this device */
	        1,				/* first configuration */
		CONFIGURATION_IDX, 		/* index string descriptor of this config */
		USB_CONFIG_SELF_POWERED | 	/* self powered and remote wake up */
		USB_CONFIG_REMOTE_WAKEUP,
	        MAXPWR_mA(2)			/* 2 mA consumption from bus */
	},{	sizeof(INTERFACE_DESCRIPTOR),	/* size of this descriptor in bytes */
		INTERFACE_DESCR,		/* interface descriptor type */
		IDX_INTERFACE_NR0, 		/* index in the array of current interfaces supported */
		ALT_INTERFACE_NR0,		/* no alternate interface */
		(EPPS-EPP_OFF)*2,		/* number of endpoints used */
		0x09,				/* USB class 9 */
		0x00,				/* no subclass codes used */
		0x00,				/* no class specific protocol on this interface */
		INTERFACE_IDX			/* index string descriptor of this interface */
	},
						/* configure bulk endpoints */
#if ( EPPS >= 1 ) && ( EPP_OFF < 1 )
		EPIN_DESCR(  1, TT_BULK, EP1_LOOP_BACK_FIFO_SIZE, 0 ),
		EPOUT_DESCR( 1, TT_BULK, EP1_LOOP_BACK_FIFO_SIZE, 0 ),
#endif
#if ( EPPS >= 2 ) && ( EPP_OFF < 2 )
		EPIN_DESCR(  2, TT_BULK, EP2_LOOP_BACK_FIFO_SIZE, 0 ),
		EPOUT_DESCR( 2, TT_BULK, EP2_LOOP_BACK_FIFO_SIZE, 0 ),
#endif
#if ( EPPS >= 3 ) && ( EPP_OFF < 3 )
		EPIN_DESCR(  3, TT_BULK, EP3_LOOP_BACK_FIFO_SIZE, 0 ),
		EPOUT_DESCR( 3, TT_BULK, EP3_LOOP_BACK_FIFO_SIZE, 0 ),
#endif
#if defined( _REG930AX_SFR ) && ( EPPS >= 4 ) && ( EPP_OFF < 4 )
		EPIN_DESCR(  4, TT_BULK, EP4_LOOP_BACK_FIFO_SIZE, 0 ),
		EPOUT_DESCR( 4, TT_BULK, EP4_LOOP_BACK_FIFO_SIZE, 0 ),
#endif
#if defined( _REG930AX_SFR ) && ( EPPS >= 5 ) && ( EPP_OFF < 5 )
		EPIN_DESCR(  5, TT_BULK, EP5_LOOP_BACK_FIFO_SIZE, 0 ),
		EPOUT_DESCR( 5, TT_BULK, EP5_LOOP_BACK_FIFO_SIZE, 0 ),
#endif
};

/* 
 *	LANGUAGE ID's supported text
 */
static unsigned char string_language[4] = 
{ 
	sizeof(string_language), STRING_DESCR, 
	0x09, 0x04 	/* 2 Byte long id for English: i.e. 0409h */
}; 

/*
 *	MANUFACTURER String Descriptor text
 */
static unsigned char string_manufacturer[44] = 
{ 
	sizeof(string_manufacturer), STRING_DESCR,
	'T',0,'a',0,'s',0,'k',0,'i',0,'n',0,'g',0,' ',0,
	'S',0,'o',0,'f',0,'t',0,'w',0,'a',0,'r',0,'e',0,' ',0,
	'B',0,'.',0,'V',0,'.',0
};

/*
 *	PRODUCT String Descriptor text 
 */
#if defined( _REG930AX_SFR ) || defined( _REG931AX_SFR )
static unsigned char string_product[52] = 
#else
static unsigned char string_product[60] = 
#endif
{ 
	sizeof(string_product), STRING_DESCR,
#if defined( _REG930AX_SFR )
	'8',0,'X',0,'9',0,'3',0,'0',0,'A',0,'X',0,' ',0,
#else 
#if defined( _REG931AX_SFR )
	'8',0,'X',0,'9',0,'3',0,'1',0,'A',0,'X',0,' ',0,
#else
#if defined( _REG930HX_SFR )
	'8',0,'X',0,'9',0,'3',0,'0',0,'H',0,'X',0,' ',0,'H',0,'u',0,'b',0,' ',0,
#else
	'8',0,'X',0,'9',0,'3',0,'1',0,'H',0,'X',0,' ',0,'H',0,'u',0,'b',0,' ',0,
#endif
#endif
#endif
	'E',0,'m',0,'b',0,'e',0,'d',0,'d',0,'e',0,'d',0,' ',0,
	'F',0,'u',0,'n',0,'c',0,'t',0,'i',0,'o',0,'n',0
};

/*
 *	SERIAL # String Descriptor text
 */
static unsigned char string_serial[54] = 
{ 
	sizeof(string_serial), STRING_DESCR,
	'C',0,'o',0,'d',0,'e',0,':',0,' ',0,
	'F',0,'U',0,'N',0,'C',0,'T',0,'I',0,'O',0,'N',0,'.',0,'C',0,' ',0,
	'R',0,'e',0,'v',0,'.',0,' ',0,'1',0,'.',0,'0',0,'0',0
};

/*
 *	CONFIGURATION String Descriptor text
 */
static unsigned char string_configuration[68]= 
{
	sizeof(string_configuration), STRING_DESCR,
	'E',0,'V',0,'A',0,'L',0,' ',0,'B',0,'o',0,'a',0,'r',0,'d',0,' ',0,'S',0,'e',0,'t',0,'u',0,'p',0,',',0,' ',0,
	('0'+(EPPS-EPP_OFF)),0,'E',0,'P',0,'P',0,' ',0,'L',0,'o',0,'o',0,'p',0,'b',0,'a',0,'c',0,'k',0,',',0,' ',0,
};

/*
 *	INTERFACE String Descriptor text 
 */
static unsigned char string_interface[54] = 
{
	sizeof(string_interface), STRING_DESCR,
	'<',0,'I',0,'n',0,'t',0,'e',0,'r',0,'f',0,'a',0,'c',0,'e',0,' ',0,
	'S',0,'t',0,'r',0,'i',0,'n',0,'g',0,' ',0,'I',0,'n',0,'d',0,'e',0,'x',0,
	' ',0,('0'+INTERFACE_IDX),0,'>',0
};

/*
 *	SOF_isr output subroutines
 */
static void out_token				( void );
static void setup_token				( void );
static void device_request			( void );
static void get_standard_device_command		( void );
static void set_standard_endpoint_command	( void );
static void set_standard_device_command		( void );
static void get_standard_endpoint_command	( void );
static void set_standard_interface_command	( void );
static void get_standard_interface_command	( void );
static void reset_fifos				( void );
static void set_standard_device_configuration	( unsigned int );

/*
 *	function_isr input subroutines
 */
static void in_token				( void );
static void control_buffer			( unsigned int, unsigned char * );

/*
 *	Initialization subroutines
 */
static void initialize_function_ep0		( void );
static void initialize_variables		( void ) {}


static COMMAND		command;		/* command control buffer */
static unsigned char 	stage;			/* read/write/setup machine stages */
static unsigned int	device_configuration;	/* device configuration value */
static unsigned char	function_RWU;		/* administrates remote wakeup status of function */

/* 
 * When a control Write with a data stage is detected
 * the data is placed in this buffer. The size of data 
 * is limited to this buffer length.
 */
static unsigned char 	control_write_data_buffer[8];
static unsigned char	control_write_data_pointer;

/*****************************************************************************
|*
|*  FUNCTION:           initialize_function_ep0
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*	EP0 Comes up as it should except that all flow control is 
|*	turned off except for OE bits. This will turn them on so we 
|*	can accept all data. All other endpoints should not be configured
|*	until a Set Configuration command is given.
|*
 */
static void initialize_function_ep0( void )
{
	unsigned char save = EPINDEX;
	
        EPINDEX = 0x00;		/* Select Function EP0 */
	EPCON	= 0x3f;		/* enable receiver and transmitter */
	EPINDEX = save;
	FIE     = 0x03;		/* Enable function endpoint 0 TX and RX interrupts. */
}

/*****************************************************************************
|*
|*  FUNCTION:           initialize_function
|*
|*  AVAILABILITY:	external
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Initialize embedded function. First initialize global variables.
|*	Call reset routine vendor code. Initialize end point 0. Enable Start
|*	of frame interrupt and Function ISR.
|*
 */
void initialize_function( void )
{
	initialize_variables();		/* Initialize the RAM space as required */
        initialize_function_ep0();	/* initialize endpoint 0 */

	SET_SOFIE;			/* Enable SOF Interrupts */
	SET_ESOF;			/* Enable SOF ISR,            IEN1.0 */
        SET_EF;				/* Enable Function ISR,       IEN1.1 */
	SET_ESR;			/* Enable Suspend/Resume ISR, IEN1.2 */
#if defined( _REG930AX_SFR ) && ( EPPS > 3 )
	SIXEPPEN = 1;			/* enable six-endpoint pairs */
#endif
}

/*****************************************************************************
|*
|*  FUNCTION:           get_standard_device_command
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Process a Get Standard Device Command Setup Token.
|*	This can only be a Get (Device, Configuration or String) 
|*	Descriptor, or Get Device Status, Get Configuration.
|*
 */
static void get_standard_device_command( void )
{
	unsigned char status;

	switch( bRequest )
	{
	case GET_CONFIGURATION:		/* get configuration */
		TXDAT  = device_configuration;
		TXCNTL = 0x01;
		break;
	case GET_STATUS:		/* get device status */
		status = 0;		/* clear remote wakeup and self powered */
		/*
		 * Check the software RWU flag to see if embedded 
		 * function remote wakeup is enabled. function_RWU is set
		 * or cleared by device command SET or CLEAR FEATURE.
		 */
		if ( config1.config.bmAttributes & USB_CONFIG_REMOTE_WAKEUP )
		{
			status |= function_RWU; /* set RWU enabled bit */
		}
		if ( config1.config.bmAttributes & USB_CONFIG_SELF_POWERED )
		{
			status |= 0x01; 	/* set self powered bit */
		}
	        TXDAT  = status;
	        TXDAT  = 0x00;
	        TXCNTL = 0x02;
		break;
	case GET_DESCRIPTOR:
		switch( bDescriptortype )
		{
		case DEVICE_DESCR:	/* get device descriptor */
		        control_buffer( sizeof(DEVICE_DESCRIPTOR), (unsigned char *)&device );
			break;
		case CONFIG_DESCR:	/* get configuration descriptor */
		        control_buffer( sizeof(CONF_DESCRIPTOR), (unsigned char *)&config1 );
			break;
		case STRING_DESCR: 	/* get string descriptor */
			/* 
			 * if string index = 0 ... regardless of language ID, 
			 * return the language ID's supported by this device
			 * 
			 * First check to make sure wIndex = language ID = 0 
			 * or 0x0409 -> English.  If not stall.
			 */
			if ( bDescriptorIndex == 0 || windex == 0x0409 )
			{
				 switch( bDescriptorIndex )
				 {
				 case LANGUAGE_IDX:  	/* return list of languages supported */
				        control_buffer( sizeof(string_language), string_language );
					break;
				/* 
				 * NEW CODE-Return English ID Codes
				 */				
				case MANUFACTURER_IDX:	/* return manufacturer string */
				        control_buffer( sizeof(string_manufacturer), string_manufacturer );
					break;
				case PRODUCT_IDX:	/* return product string */
				        control_buffer( sizeof(string_product), string_product );
					break;
				case SERIAL_IDX:  	/* return serial number string */
				        control_buffer( sizeof(string_serial), string_serial );
					break;
				case CONFIGURATION_IDX:	 /* return configuration string */
				        control_buffer( sizeof(string_configuration), string_configuration );
					break;
				case INTERFACE_IDX:  	/* return interface string */
				        control_buffer( sizeof(string_interface), string_interface );
					break;
				default:
					goto stall_ep0;
//					break;
				}
			}
			else
			{
				goto stall_ep0;
			}
			break;
		default:
			goto stall_ep0;
//			break;
		}
		break;
	default:
stall_ep0:
		EPCON |= (SET_IN_STALL_MASK | SET_OUT_STALL_MASK);	/* stall EP0 */
		break;
	}
}

/*****************************************************************************
|*
|*  FUNCTION:           reset_fifos
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      This routine flushes and clears all of the fifos.
|*	The routine also sets data toggle back to zero as
|*	defined in the USB spec. This routine would be called
|*	when ever you change configurations.
|*
 */
static void reset_fifos( void )
{
        unsigned char fifo;	/* Number of FIFOs */

#if ( EPPS > EPP_OFF )
	FIE   &= 0x03;		/* Disable interrupts for endpoint 1-3 */
        FIFLG &= 0x03;		/* Disable interrupts for function 1-3 */
#if defined( _REG930AX_SFR ) && ( EPPS > 3 )
	FIE1   = 0x00;		/* Disable interrupts for endpoint 4-5 */
        FIFLG1 = 0x00;		/* Disable interrupts for function 1-3 */
#endif
#endif

	for ( fifo = (1+EPP_OFF); fifo <= EPPS; fifo++ )
	{
	        EPINDEX = fifo;

		SET_TXCLR;	/* Flush the TX FIFOs */
		SET_RXCLR;	/* Flush the RX FIFOs */

		/* 
		 * Clear TXSEQ, set TXSOVW and reset data
		 * toggle back to zero
		 */
	        TXSTAT = (TXSTAT & 0x7F) | 0x08;

		/* 
		 * Clear RXSEQ, set RXSOVW and reset data 
		 * toggle back to zero 
		 */
	        RXSTAT = (RXSTAT & 0x7F) | 0x08;

		EPCON = 0x10;	/* Reset value, receive single packed mode */
	}
}

/*****************************************************************************
|*
|*  FUNCTION:           set_standard_device_configuration
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*	config	USB configuration number
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Setup the endpoints to the parameters specified 
|*	in the configuration descriptor specified.
|*
 */
static void set_standard_device_configuration( unsigned int config )
{
        unsigned char fifo;
	unsigned char save = EPINDEX;


	switch( config )
	{
	case 0:			    	/* Config 0 */
		/* called when device is disabled in
		 * hardware profile from the device manager 
		 */
	        reset_fifos();	    	/* Place all FIFOs in known state */
		goto setup_null_packet;
	case 1: 		    	/* Config 1 */
	        reset_fifos();	    	/* Place all FIFOs in known state */
		for ( fifo = (1+EPP_OFF); fifo <= EPPS; fifo++ )
		{
			unsigned char imask = 0x3 << ((fifo & 0x3) * 2);

			EPINDEX = fifo;		/* select endpoint */
			EPCON   = 0x0F;		/* Non Control, Non ISO, Dual Packet */
			/* 
			 * Enable fifo the interrupts 
			 */
#if defined( _REG930AX_SFR ) && ( EPPS > 3 )
			if ( fifo > 3 )
			{
				FIE1 |= imask;
			}
			else
#endif
			{
				FIE |= imask;
			}
		}
setup_null_packet:
		EPINDEX = save;
		TXCNTL = 0x0;	    	/* Setup Null Packet */
		break;
	default:			/* bad set standard device configuration */
		EPCON |= (SET_IN_STALL_MASK | SET_OUT_STALL_MASK);	/* stall EP0 */
		break;
	}

	device_configuration = config;	/* Set up the variables */
}

/*****************************************************************************
|*
|*  FUNCTION:           set_standard_device_command
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Process a Standard Set Device Command Setup Token
|*	The only valid commands are Set Address, Set Configuration, 
|*	Clear Feature and Set Feature. Other commands which have this 
|*	field but are un-defined are Set Descriptor.
|*
 */
static void set_standard_device_command( void )
{
	switch( bRequest )
	{
	case SET_CONFIGURATION:	/* set device configuration */
		set_standard_device_configuration( wvalue );
		break;
	case SET_ADDRESS:	/* set device address */
                /*
		 * Set address should just return.  
		 * The command will be executed following the status 
		 * stage of control write. The USB device does not change 
		 * its device address until after the status stage of this
		 * request is completely successfully.
		 */
		TXCNTL = 0x0;	/* Setup Null Packet */
		break;
	case SET_FEATURE:	/* Set device RWU */
		if( wvalue == DEVICE_REMOTE_WAKEUP )
		{
			/*	
			 * Set sw flag for RWU enabled 
			 * Set the embedded function remote wakeup 
			 * flag to 02 setting status.1 
			 */
			function_RWU = 0x02;
			TXCNTL = 0x0;	/* Setup Null Packet */
		}
		else
		{
			goto stall_ep0;
		}
		break;
	case CLEAR_FEATURE:	/* Clear Device RWU */
		if ( wvalue == DEVICE_REMOTE_WAKEUP )
		{
			/* 
			 * Clear sw flag for RWU disabled 
			 * Clear the end function remote wakeup 
			 * flag to zero
			 */
			function_RWU = 0x00;
			TXCNTL = 0x0;	/* Setup Null Packet */
		}
		else
		{
			goto stall_ep0;
		}
		break;
	case SET_DESCRIPTOR:
		/* 
		 * Optional, if supported the descriptor may need
		 * updated or a new descriptor may be added.
		 */
	default:		/* Bad set standard device command */
stall_ep0:
		EPCON |= (SET_IN_STALL_MASK | SET_OUT_STALL_MASK);	/* stall EP0 */
		break;
	}
}

/*****************************************************************************
|*
|*  FUNCTION:           get_standard_endpoint_command
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Process a Standard Get Endpoint Command Setup Token
|*	This can only be a get status or synch frame.
|*
 */
static void get_standard_endpoint_command( void )
{
	unsigned char save;
 	unsigned char stall = 0;	 	/* assume stall clear */

	switch( bRequest )
	{
	case GET_STATUS:			/* get endpoint status */
		save = EPINDEX;			/* save endpoint index */
		/* 
		 * set the endpoint to get status from 
		 */
		EPINDEX = (save & 0x80)|(windex & 0xF);	

		if ( USB_DEVICE_TO_HOST( windex ) )	/* get in stall status */
		{
			if ( GET_TXSTL )
			{
				stall = 1;	/* return stalled transmitter */
			}
		}
		else				/* get out stall status */
		{
			if ( GET_RXSTL )
			{
				stall = 1;	/* return stalled receiver */
			}
		}
	        EPINDEX  = save;		/* Point the index back to EP0 */
	        TXDAT    = stall;		/* return stall status */
	        TXDAT    = 0x00;
	        TXCNTL   = 0x02;
		break;
	case SYNCH_FRAME:			/* synch endpoint */
		/*
		 * Only supported for isochronous transfers
		 */
	default:				/* Bad get standard endpoint command */
		EPCON |= (SET_IN_STALL_MASK | SET_OUT_STALL_MASK);	/* stall EP0 */
		break;
	}
}

/*****************************************************************************
|*
|*  FUNCTION:           set_standard_endpoint_command
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Process a Standard Set Endpoint Command Setup Token.
|*	This can only be a Set or Clear Feature Endpoint Stall
|*
 */
static void set_standard_endpoint_command( void )
{
	unsigned char save = EPINDEX;

	/* 
	 * Get the endpoint of the stall to clear 
	 */
	unsigned char index = (windex & 0xF);

	/*
	 * Setup the EPINDEX to point at correct index.
	 */
	EPINDEX = (save & 0x80) | index;

	/* 
	 * Check to make sure this is an Endpoint Stall.
	 */
	if ( wvalue == ENDPOINT_STALL )
	{
	        switch( bRequest )
		{
	        case CLEAR_FEATURE: 	/* clear endpoint stall */
			if ( index )	/* EP1..EPn Clear Stall Command ? */
			{
				if ( USB_DEVICE_TO_HOST( windex ) )
				{
					/* 
					 * Reset the input pipe ( host view )
					 */
					EPCON &= CLEAR_IN_STALL_MASK;
				}
				else
				{
					/* 
					 * Reset the output pipe ( host view )
					 */
					EPCON &= CLEAR_OUT_STALL_MASK;
				}
			}
			else		/* EP0 Clear Stall Command */
			{
				/* 
				 * For EP0 Clear both TX & RX stall bits 
				 */
				EPCON &= (CLEAR_IN_STALL_MASK & CLEAR_OUT_STALL_MASK);
			}
			break;
		case SET_FEATURE:	/* set endpoint stall */
			if ( index )	/* EP1..EPn Set Stall Command ? */
			{
				if ( USB_DEVICE_TO_HOST( windex ) )
				{
					EPCON |= SET_IN_STALL_MASK;
				}
				else
				{
					EPCON |= SET_OUT_STALL_MASK;
				}
			}
			else		/* EP0 set stall command */
			{
				goto stall_ep0;	/* just simply stall EP 0 */
			}
			break;
		default:		/* Undefined set standard endpoint command */
			goto stall_ep0;
//			break;
		}
		EPINDEX = save;		/* restore endpoint index */
		TXCNTL  = 0x0;		/* Setup Null Packet */
	}
	else
	{
stall_ep0:
		EPINDEX = save;		/* restore endpoint index */
		EPCON |= (SET_IN_STALL_MASK | SET_OUT_STALL_MASK);	/* stall EP0 */
	}
}

/*****************************************************************************
|*
|*  FUNCTION:           get_standard_interface_command
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      This request allows the host to determine the currently selected 
|*	alternative setting of the interface.
|*
 */
static void get_standard_interface_command( void )
{
	if ( bRequest == GET_INTERFACE	&&
	     windex == IDX_INTERFACE_NR0 )
	{
		TXDAT = ALT_INTERFACE_NR0;
	        TXCNTL = 0X01;
	}
	else
	{
		EPCON |= (SET_IN_STALL_MASK | SET_OUT_STALL_MASK);	/* stall EP0 */
	}
}

/*****************************************************************************
|*
|*  FUNCTION:           set_standard_interface_command
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      This request allows the host to select an alternate setting for 
|*	the specified interface. 
|*
 */
static void set_standard_interface_command( void )
{
	if ( bRequest == SET_INTERFACE		&&
	     wvalue == ALT_INTERFACE_NR0 	&&
	     windex == IDX_INTERFACE_NR0	)
	{
		TXCNTL = 0x0;		/* Setup Null Packet */
	}
	else
	{
		EPCON |= (SET_IN_STALL_MASK | SET_OUT_STALL_MASK);	/* stall EP0 */
	}
}

/*****************************************************************************
|*
|*  FUNCTION:           device_request
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Determine USB Device request from the host on the device's default 
|*	pipe. The request type of the setup packet is processed. For invalid
|*	or reserved commands a stall is generated.
|*
 */
static void device_request( void )
{
	switch( bmRequestType & 0xE3 )
	{
	case	GET_STANDARD_DEVICE_COMMAND:
		get_standard_device_command();
		break;
	case	SET_STANDARD_DEVICE_COMMAND:
		set_standard_device_command();
		break;
	case	GET_STANDARD_ENDPOINT_COMMAND:
		get_standard_endpoint_command();
		break;
	case	SET_STANDARD_ENDPOINT_COMMAND:
		set_standard_endpoint_command();	
		break;
	case	SET_STANDARD_INTERFACE_COMMAND:
		set_standard_interface_command();
		break;
	case	GET_STANDARD_INTERFACE_COMMAND:
		get_standard_interface_command();
		break;

	case	SET_VENDOR_DEVICE_COMMAND:     
	case	GET_VENDOR_DEVICE_COMMAND:     

	case	GET_VENDOR_INTERFACE_COMMAND:
	case	SET_VENDOR_INTERFACE_COMMAND:
	case	SET_VENDOR_ENDPOINT_COMMAND:
	case	GET_VENDOR_ENDPOINT_COMMAND:
	case	SET_VENDOR_OTHER_COMMAND:		
	case	GET_VENDOR_OTHER_COMMAND:
		/*
	         * UnknownVendorCommand -- STALL ENDPOINT
		 */
		/* FALL THROUGH */

	case	SET_STANDARD_OTHER_COMMAND:
	case	GET_STANDARD_OTHER_COMMAND:
		/*
	         * UnknownVendorCommand -- STALL ENDPOINT
		 */
		/* FALL THROUGH */

	case	GET_CLASS_DEVICE_COMMAND:
	case	GET_CLASS_OTHER_COMMAND:
	case	SET_CLASS_DEVICE_COMMAND:
	case	SET_CLASS_ENDPOINT_COMMAND:
	case	SET_CLASS_OTHER_COMMAND:
	case	SET_CLASS_INTERFACE_COMMAND:
	case	GET_CLASS_INTERFACE_COMMAND:
	case	GET_CLASS_ENDPOINT_COMMAND:
		/*
	         * UnknownVendorCommand -- STALL ENDPOINT
		 */
		/* FALL THROUGH */

	default:			/* ReservedCommand */
		EPCON |= (SET_IN_STALL_MASK | SET_OUT_STALL_MASK);	/* stall EP0 */
		break;
	}
}

/*****************************************************************************
|*
|*  FUNCTION:           setup_token
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Process a Setup token received on end point 0.
|*
|*	The First step is setting up the Stage tracing variable by examine 
|*	the bmRequest value. This will tell us if this is a control Write
|*	or control read command.
 */
static void setup_token( void )
{
	FRXD0 = 0;		/* Clear the interrupt bit. */

	/*
	 * Now Check for Data still in FIFO.
	 * If this is true then the function
	 * did not see the ACK form the host.
	 */
	 if ( TXFLG & 0xC0 )
	 {
		/* Flush the Transmit FIFOS in case a null packet is still left. */
	 	SET_TXCLR;
		TXCNTL = 0x00;	/* OHCI fix for device Errata */
		SET_TXCLR;	/* OHCI fix for device Errata */
	}

	if ( EPCON & (SET_IN_STALL_MASK | SET_OUT_STALL_MASK) )		/* Stall EP0 ? */
	{
		EPCON  &= ~(SET_IN_STALL_MASK | SET_OUT_STALL_MASK);	/* Clear the stalls */
		TXSTAT |= 0x88;	/* Set data toggle on TX ep0. */
	}

	/*
	 * 8 is length of all setup packets.
	 */
	if ( RXCNTL == 8 )
	{
		/*
		 * Move FIFO to command buffer.
		 */
	       	CLR_EDOVW;

	       	bmRequestType	 = RXDAT;
		bRequest	 = RXDAT;
		bDescriptorIndex = RXDAT;	/* wValue LSB */
		bDescriptortype	 = RXDAT;	/* wValue MSB */
		blsbIndex	 = RXDAT;
		bmsbIndex	 = RXDAT;
		blsbLength	 = RXDAT;
		bmsbLength	 = RXDAT;

	        SET_RXFFRC;	        	/* Update receive FIFO state */

		if ( USB_DEVICE_TO_HOST( bmRequestType ) )	/* Setup Get Command ? */
		{
			stage = READ_STAGE;	/* Advance State Machine to control read stage */
		}
		else				/* Setup Set Command */
		{
			stage = WRITE_STAGE;	/* Advance State Machine to control write state */

			/*
	                 * Now check to see if this is a control write with a data stage.
			 */
			if ( wlength != 0 )
			{
				/* 
				 * Initialize the data pointer for future 
				 * control write data stages. 
				 */
				control_write_data_pointer = 0;

				/* 
				 * If we are expecting data then don't
				 * process the command yet. Exit and wait
				 * for the rest of the data to come in.
				 * Command is processed at end of control
				 * write data stage. See out_token
				 * WRITE_STAGE
				 */
				goto rx_setup;
			}
		}

		device_request();
	}

rx_setup:
        CLR_RXSETUP;
}

/*****************************************************************************
|*
|*  FUNCTION:           out_token
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Services all OUT tokens on end point 0.
|*
 */
static void out_token( void )
{
	unsigned char count;

	FRXD0 = 0;			/* Clear the interrupt bit. */

	switch( stage )
	{
	case READ_STAGE:		/* Status stage of control read */
		SET_RXFFRC;		/* Update receive FIFO state */
		stage = SETUP_STAGE;	/* Advance State Machine to setup stage */
		break;
	case WRITE_STAGE:		/* Data stage of control write, e.g. set descriptor */
		count = RXCNTL;		/* Get number of bytes in receiver fifo */
		while( count-- )	/* Get bytes from the FIFO */
		{
			control_write_data_buffer[control_write_data_pointer++] = RXDAT;
		}
		SET_RXFFRC;  		/* Update receive FIFO state */
		/*
		 * check to see if this was the last read
		 */
		if ( control_write_data_pointer == blsbLength )
		{
			 /* 
			  * At this point all data has been 
                          * stored and all we need to do now is to
			  * process it.
			  */
			 device_request();
		}	/* else need more data */
		break;
	default:
		EPCON |= (SET_IN_STALL_MASK | SET_OUT_STALL_MASK);	/* stall EP0 */
		break;
	}
}

/*****************************************************************************
|*
|*  FUNCTION:           in_token
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Services all IN tokens on end point 0.
|*
 */
static void in_token( void )
{
	FTXD0 = 0;		/* clear the interrupt bit */

	switch( stage )
	{
	case WRITE_STAGE:	/* Status stage of control write */
		if ( bRequest == SET_ADDRESS )
		{
			/* 
			 * Complete standard set address command
			 */
			FADDR = wvalue;
		}
		stage = SETUP_STAGE;	/* Advance State Machine to setup stage */
		break;
	case READ_STAGE:	/* Data stage of control read */
		control_buffer( 0, (unsigned char *)0 );
		break;
	default:
		EPCON |= (SET_IN_STALL_MASK | SET_OUT_STALL_MASK);	/* stall EP0 */
		break;
	}
}

/*****************************************************************************
|*
|*  FUNCTION:           control_buffer
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*	size	buffer length in bytes
|*	p	pointer to buffer
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Set global buffer pointer, initialize the buffer counter and
|*	copy data from the location pointed to by p.
|*
|*	The number of bytes left to transmit are stored in buffer_size.
|*
|*	If zero bytes are left then a Null packet is loaded even if the 
|*	null/short packet is not needed.
|*
 */
static void control_buffer( unsigned int size, unsigned char * p )
{
	unsigned char		count;		/* Number of bytes in FIFO Always <=16 */
	static unsigned char *  buffer;
	static unsigned int	buffer_size;


	if ( p )				/* Transmit new buffer ? */
	{
		 /*
		  * Only transmit the amount asked for or available
		  */ 
		if ( size > wlength )
		{
	        	size = wlength;		/* amount asked for is smaller then available */
		}
	}				     
	else					/* continue transmitting current buffer */
	{
		p    = buffer;			/* Get current location */
		size = buffer_size;		/* Get byte left over */
	}

	for ( count = 0; count < EP0_MAX_PACKET_SIZE && size; count++, size-- )
	{
		TXDAT = p[count];		/* Transmit the data */
	}

	buffer      = p+count;			/* Update buffer pointer   */
	buffer_size = size;			/* Update buffer size left */

        TXCNTL = count;				/* Write count into TXCNT register */
	SET_TXOE;				/* Enable data transmit */
}

/*****************************************************************************
|*
|*  FUNCTION:           SOF_isr
|*
|*  AVAILABILITY:	extern
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Service the Start Of Frame Interrupt
|*
|*	This routine simply displays the upper three bytes of the
|*	SOF in the lower three bits of the LED's on P1.
|*	It does not affect the other LED's. Very useful
|*	in determining when the function is receiving SOFs
|*
 */
#if defined( _REG930AX_SFR ) || defined( _REG931AX_SFR )
void _interrupt( 8 ) _using( 0 ) SOF_isr( void )
{
	if ( GET_ASOF )	/* Any Start of Frame */
	{
		CLR_ASOF;

		P1_0 = GET_TS8;
		P1_1 = GET_TS9;
		P1_2 = GET_TS10;
	}
}
#endif

/*****************************************************************************
|*
|*  FUNCTION:           function_isr
|*
|*  AVAILABILITY:	extern
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Service the function_isr Interrupt
|*
|*	This routine scans the interrupt pending bits in order
|*	and branches to any routines that have pending interrupts.
|*	After finishing the routine jumps back to the top of
|*	the scanning loop to preserve interrupt priority.
|*
|*	NOTE: The order of precedence is set by the order in which the
|*	user places the checks.
|*
 */
void _interrupt( 9 ) _using( 0 ) function_isr( void )
{
	unsigned char save = EPINDEX;

	do
	{
		/* NOTE: Priority of the check sequence determines priority. */

		/*
		 * First check to see if this is EP0 Command
		 */
		while ( FTXD0 || FRXD0 )	
		{
			/* 
			 * Process the EP0 command in the correct order 
			 *
			 * If this an EP0 command then we need to process TX & RX
			 * ISRs in the correct order if, for whatever reason they 
			 * occurred back to back. This happens in OHCI systems.
			 */
			EPINDEX = 0x00;

			/* Now check to see what stage the TX/RX is in.
                         * If it's in the control read stage then check TX first.
                         * If it's in the setup or control write stage then check the RX first.
			 */
			if ( stage == READ_STAGE )
			{
				/*
				 * Process the TX FIFO first
                                 */
				if ( FTXD0 )
				{
				        in_token();	/* 8293X has sent a packet to HC */
				}
				else /* if ( FRXD0 ) */
				{
					if ( GET_RXSETUP )		/* Is this a setup packet ? */
					{
						setup_token();
					}
					else
					{
					        out_token();	/* 8293X has received a packet from HC */
					}
				}
			}
			else	/* stage is SETUP_STAGE or WRITE_STAGE */
			{
				/*
				 * Process the RX FIFO first
				 */
				if ( FRXD0 )
				{
					if ( GET_RXSETUP )		/* Is this a setup packet ? */
					{
						setup_token();
					}
					else
					{
					        out_token();	/* 8293X has received a packet from HC */
					}
				}
				else /* if ( FTXD0 ) */
				{
				        in_token();	/* 8293X has sent a packet to HC */
				}
			}
		        /* Look at all Interrupt flags again */
		}
#if ( EPPS > EPP_OFF  )
		loopback();
#endif
#if defined( _REG930AX_SFR ) && ( EPPS >= 4 )
	} while ( FIFLG && FIFLG1 && (EPPS-EPP_OFF) );	/* Check to see if any other Interrupts are pending. */
#else  
	} while ( FIFLG && (EPPS-EPP_OFF) );		/* Check to see if any other Interrupts are pending. */
#endif

	EPINDEX = save;
}

#ifdef INTERNAL_PORT
/*****************************************************************************
|*
|*  FUNCTION:           suspend_resume_isr
|*
|*  AVAILABILITY:	external
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Service the suspend and resume interrupt service routine.
|*
 */
void _interrupt( 10 ) _using( 0 ) suspend_resume_isr( void )
{
	SET_PD;		/* power down */


	/*
	 * Let device know that a resume signal was detected.
	 */
	CLR_GRSM;	/* Clear resume flag */
	CLR_GSUS;	/* Clear suspend flag */
}
#endif
