/*****************************************************************************
|*
|*  VERSION CONTROL:    @(#)hub.c	1.1 98/06/19
|*
|*  IN PACKAGE:         HX93X Firmware
|*
|*  COPYRIGHT:          Copyright (c) 1998 Tasking Software B.V.
|*
|*  DESCRIPTION:
|*
|*	Universal serial bus HX93X hub class device, hub endpoint 0
|*	and other hub firmware code
|*
 ****************************************************************************/

#include "eval.h"
#include "usb.h"
#include "hub.h"
#include "function.h"

#if !defined( _C251 ) && !defined( _CC51 )
#define _using( X )
#endif

/*
 *	Hub descriptor
 */
HUB_DESCRIPTOR hub = 
{
	sizeof(HUB_DESCRIPTOR),	/* size of this descriptor in bytes */
	HUB_DESCR,		/* device descriptor type */	
	DOWN_STREAMS,		/* number of down streams including 1 embedded stream */
#ifdef INTERNAL_PORT
	SWAP16(0x0004),		/* Hub characteristics: ganged power, compound, global over-current proctection */
#else
	SWAP16(0x0000),		/* Hub characteristics: ganged power, not compound, global over-current proctection */
#endif
	PWRON2PWRGOOD(44), 	/* approx 44 msec for power to stabilize */
	100,			/* maximum current requirements is 100 milliamps */
	1<<EMBEDDED_PORT_NUMBER,/* embedded port 4 is not removable */
	0			/* all ports respond to ganged power commands */
};

/*
 *	Config Descriptor structure
 */
struct Conf
{
	CONFIG_DESCRIPTOR	config;
	INTERFACE_DESCRIPTOR	interface;
	ENDPOINT_DESCRIPTOR	ep;
};

typedef	struct	Conf		CONF_DESCRIPTOR;

/*
 *	Descriptors
 */
static DEVICE_DESCRIPTOR device = 
{
	sizeof(DEVICE_DESCRIPTOR),		/* size of this descriptor in bytes */
        DEVICE_DESCR,				/* device descriptor type */
	SWAP16(USB_SPEC), 			/* version 1.00 compliant */
        9,					/* Hub class specification 9 */
	0,					/* with own class information */
	0,					/* no class specific protocols */
	EP0_MAX_PACKET_SIZE,			/* packet size 8 byte max for EP0 */
	SWAP16(0x042B),				/* Vendor ID */
	SWAP16(0x9304),				/* product ID */
	SWAP16(0x0150),				/* device ID */
	MANUFACTURER_IDX, 	   		/* index string descriptor manufacturer */
	PRODUCT_IDX, 	   			/* index string descriptor product */
	0,//SERIAL_IDX,		 		/* index string serial number, No serial number */
	1					/* one configuration */
};

/*
 *	Initialize Config 1 descriptor 
 */
static CONF_DESCRIPTOR config1 =
{	{ 	sizeof(CONFIG_DESCRIPTOR),	/* size of this descriptor in bytes */
		CONFIG_DESCR,			/* configuration descriptor type */
		SWAP16(sizeof(CONF_DESCRIPTOR)),/* combined descriptor size */
		1,				/* one interface supported by Hub */
	        1,				/* first configuration */
		CONFIGURATION_IDX,		/* index string descriptor of this config */
		USB_CONFIG_SELF_POWERED | 	/* self powered and remote wake up */
		USB_CONFIG_REMOTE_WAKEUP,
	        MAXPWR_mA(2)			/* 2 mA consumption from bus */
	},{	sizeof(INTERFACE_DESCRIPTOR),	/* size of this descriptor in bytes */
		INTERFACE_DESCR,		/* interface descriptor type */
		IDX_INTERFACE_NR0,		/* index in the array of current interfaces supported */
		ALT_INTERFACE_NR0,		/* no alternate interface */
		1,				/* one endpoint used on Hub */
		9,				/* Hub class specification 9 */
		0,				/* no subclass codes used */
		0,				/* no class specific protocol on this interface */
		INTERFACE_IDX			/* index string descriptor of this interface */
	},
	EPIN_DESCR( 1, TT_INTERRUPT, 1, 32 ),	/* 32 millisecond polling rate. */
};

/* 
 *	LANGUAGE ID's supported text
 */
static unsigned char string_language[4] = 
{ 
	sizeof(string_language), STRING_DESCR, 
	0x09, 0x04 	/* 2 Byte long id for English: i.e. 0409h */
}; 

/*
 *	MANUFACTURER String Descriptor text 
 */
static unsigned char string_manufacturer[44] = 
{ 
	sizeof(string_manufacturer), STRING_DESCR,
	'T',0,'a',0,'s',0,'k',0,'i',0,'n',0,'g',0,' ',0,
	'S',0,'o',0,'f',0,'t',0,'w',0,'a',0,'r',0,'e',0,' ',0,
	'B',0,'.',0,'V',0,'.',0
};

/*
 *	PRODUCT String Descriptor text 
 */
static unsigned char string_product[76] = 
{ 
	sizeof(string_product), STRING_DESCR,
#if defined( _REG930AX_SFR ) || defined( _REG930HX_SFR)
	'8',0,'X',0,'9',0,'3',0,'0',0,'H',0,'X',0,' ',0,'H',0,'u',0,'b',0,' ',0,
#else
	'8',0,'X',0,'9',0,'3',0,'1',0,'H',0,'X',0,' ',0,'H',0,'u',0,'b',0,' ',0,
#endif
	',',0,' ',0,('0'+DOWN_STREAMS),0,' ',0,'D',0,'S',0,' ',0,'P',0,'o',0,'r',0,'t',0,'s',0,
	' ',0,'(',0,'1',0,' ',0,'I',0,'n',0,'t',0,'e',0,'r',0,'n',0,'a',0,'l',0,')',0
};

/*
 *	SERIAL # String Descriptor text 
 */
static unsigned char string_serial[44] = 
{ 
	sizeof(string_serial), STRING_DESCR,
	'C',0,'o',0,'d',0,'e',0,':',0,' ',0,
	'H',0,'U',0,'B',0,'.',0,'C',0,' ',0,
	'R',0,'e',0,'v',0,'.',0,' ',0,'1',0,'.',0,'0',0,'0',0
};

/*
 *	CONFIGURATION String Descriptor text
 */
static unsigned char string_configuration[90] = 
{ 	
	sizeof(string_configuration), STRING_DESCR,
	'E',0,'V',0,'A',0,'L',0,' ',0,'B',0,'o',0,'a',0,'r',0,'d',0,' ',0,'S',0,'e',0,'t',0,'u',0,'p',0,',',0,' ',0,
	'E',0,'m',0,'b',0,'e',0,'d',0,'d',0,'e',0,'d',0,' ',0,
	'F',0,'u',0,'n',0,'c',0,'t',0,'i',0,'o',0,'n',0,' ',0,
#ifdef INTERNAL_PORT
	'E',0,'n',0,'a',0,'b',0,'l',0,'e',0,'d',0,' ',0
#else
	'D',0,'i',0,'s',0,'a',0,'b',0,'l',0,'e',0,'d',0
#endif
};

/*
 *	INTERFACE String Descriptor text 
 */
static unsigned char string_interface[54] = 
{
	sizeof(string_interface), STRING_DESCR,
	'<',0,'I',0,'n',0,'t',0,'e',0,'r',0,'f',0,'a',0,'c',0,'e',0,' ',0,
	'S',0,'t',0,'r',0,'i',0,'n',0,'g',0,' ',0,'I',0,'n',0,'d',0,'e',0,'x',0,
	' ',0,('0'+INTERFACE_IDX),0,'>',0
};

/*
 *	SOF/HUB interrupt OUT token subroutines
 */
static void out_token				( void );
static void setup_token				( void );
static void device_request			( void );

/*
 *	SOF/HUB interrupt IN token subroutines
 */
static void in_token				( void );
static void control_buffer			( unsigned int, unsigned char * );

/*
 *	Initialization subroutines
 */
static void initialize_hub_ep0			( void );
static void initialize_variables		( void ) {}

/*
 *	Hub standard command subroutines
 */
static void get_standard_device_command		( void );
static void set_standard_endpoint_command	( void );
static void set_standard_device_command		( void );
static void get_standard_endpoint_command	( void );
static void set_standard_interface_command	( void );
static void get_standard_interface_command	( void );
static void set_standard_device_configuration	( unsigned char );

/*
 *	Hub class command subroutines
 */
static void get_class_device_command		( void );
static void get_class_other_command		( void );
static void set_class_device_command		( void );
static void set_class_other_command		( void );

static void resume				( void ) {}	/* embedded function resume */
static void suspend				( void ) {}	/* embedded function suspend */

static COMMAND		command;		/* command control buffer */
static unsigned char 	stage;			/* read/write/setup machine stages */
static unsigned int	device_configuration;	/* device configuration value */

/*****************************************************************************
|*
|*  FUNCTION:           initialize_hub_ep0
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*	Initialize Hub endpoint 0. Select endpoint 0, enable the receiver and 
|*	transmitter of this endpoint. The endpoint 0 default state does not except 
|*	flow control. This function will turn them on so we can accept all data. 
|*	All other endpoints should not be configured until a Set Configuration 
|*	command is given.
|*
 */
static void initialize_hub_ep0( void )
{
	EPINDEX	= 0x80;		/* select Hub EP0, equals startup value */
	EPCON	= 0x3f;		/* enable receiver and transmitter */
	HIE	= 0x03;		/* enable the Hub EP0 interrupts */
}

/*****************************************************************************
|*
|*  FUNCTION:           initialize_hub
|*
|*  AVAILABILITY:	external
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Initialize the Hub. First initialize global variables, when needed.
|*	Initialize endpoint 0. Enable Start Of Frame, Hub Function ISR and 
|*	suspend/resume interrupts. Set connect status of embedded port when 
|*	internal port is enabled for function support.
|*
 */
void initialize_hub( void )
{
	initialize_variables();		/* Initialize the RAM space as required */
        initialize_hub_ep0();		/* initialize endpoint 0 */

	SET_ESOF;			/* Enable HUB/SOF ISR, setb IEN1.0 */
	SET_ESR;			/* Enable Suspend/Resume ISR, setb IEN1.2 */

#ifdef INTERNAL_PORT
	HPINDEX	= EMBEDDED_PORT_NUMBER;	/* set embedded port number */
	SET_PCSC;			/* connect status change */
	HPINDEX	= 0;			/* restore reset state */
#endif
}

/*****************************************************************************
|*
|*  FUNCTION:           get_standard_device_command
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Process Get Standard Device Command Setup Tokens.
|*	This can only be a Get (Device, Configuration or String) 
|*	Descriptor, Get Device Status or Get Configuration.
|*
 */
static void get_standard_device_command( void )
{
	unsigned char status;

	switch( bRequest )
	{
	case GET_CONFIGURATION:		/* get configuration */
		TXDAT  = device_configuration;
		TXCNTL = 0x01;
		break;
	case GET_STATUS:		/* get device status */
		status = 0;		/* clear remote wakeup and self powered */
		if ( config1.config.bmAttributes & USB_CONFIG_SELF_POWERED )
		{
			status |= 0x01; /* set self powered bit */
		}
		if ( GET_HRWUPE )
		{
			status |= 0x02;	/* set RWU enabled bit */
		}
	        TXDAT  = status;
	        TXDAT  = 0x00;
	        TXCNTL = 0x02;
		break;
	case GET_DESCRIPTOR:		/* get descriptor */
		switch( bDescriptortype )
		{
		case DEVICE_DESCR:	/* get device descriptor */
		        control_buffer( sizeof(DEVICE_DESCRIPTOR), (unsigned char *)&device );
			break;
		case CONFIG_DESCR:	/* get configuration descriptor */
		        control_buffer( sizeof(CONF_DESCRIPTOR), (unsigned char *)&config1 );
			break;
		case STRING_DESCR: 	/* get string descriptor */
			/* 
			 * if string index = 0 ... regardless of language ID, 
			 * return the language ID's supported by this device
			 * 
			 * First check to make sure wIndex = language ID = 0 
			 * or 0x0409 -> English.  If not stall.
			 */
			if ( bDescriptorIndex == 0 || windex == 0x0409 )
			{
				 switch( bDescriptorIndex )
				 {
				 case LANGUAGE_IDX:  	/* return list of languages supported */
				        control_buffer( sizeof(string_language), string_language );
					break;
				/* 
				 * NEW CODE-Return English ID Codes
				 */				
				case MANUFACTURER_IDX:	/* return manufacturer string */
				        control_buffer( sizeof(string_manufacturer), string_manufacturer );
					break;
				case PRODUCT_IDX:	/* return product string */
				        control_buffer( sizeof(string_product), string_product );
					break;
				case SERIAL_IDX:  	/* return serial number string */
				        control_buffer( sizeof(string_serial), string_serial );
					break;
				case CONFIGURATION_IDX:	 /* return configuration string */
				        control_buffer( sizeof(string_configuration), string_configuration );
					break;
				case INTERFACE_IDX:  	/* return interface string */
				        control_buffer( sizeof(string_interface), string_interface );
					break;
				default:
					goto stall_ep0;
//					break;
				}
			}
			else
			{
				goto stall_ep0;
			}
			break;
		default:
			goto stall_ep0;
//			break;
		}
		break;
	default:
stall_ep0:
		EPCON |= (SET_IN_STALL_MASK | SET_OUT_STALL_MASK);	/* stall EP0 */
		break;
	}
}

/*****************************************************************************
|*
|*  FUNCTION:           set_standard_device_configuration
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*	config	device configuration number
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Setup the endpoints to the parameters specified 
|*	in the configuration descriptor specified.
|*
|*	Enable or disable Hub endpoint 1. Endpoint 1 is used
|*	to inform the host of a hub or port status change. Only
|*	one configuration is available for the Hub endpoint 1.
|*
 */
static void set_standard_device_configuration( unsigned char config )
{
	switch( config )
	{
	case 0:			    		/* set not configured */
		CLR_EP1EN; 			/* Disable EP1 */
		TXCNTL = 0x0;	    		/* Setup Null Packet */
		break;
	case 1: 		    		/* set configuration 1 */
		SET_EP1EN;			/* Enable EP1 */
		TXCNTL = 0x0;	    		/* Setup Null Packet */
		break;
	default:				/* bad set standard device configuration */
		EPCON |= (SET_IN_STALL_MASK | SET_OUT_STALL_MASK);	/* stall EP0 */
		break;
	}

	device_configuration = config;		/* Set up the variables */
}

/*****************************************************************************
|*
|*  FUNCTION:           set_standard_device_command
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Process a Standard Set Device Command Setup Token.
|*	The only valid commands are Set Address, Set Configuration, 
|*	Clear Feature and Set Feature. Other commands which have this 
|*	field but are un-defined are Set Descriptor.
|*
 */
static void set_standard_device_command( void )
{
	switch( bRequest )
	{
	case SET_CONFIGURATION:	/* set device configuration */
		set_standard_device_configuration( wvalue );
		break;
	case SET_ADDRESS:	/* set device address */
                /*
		 * Set address should just return.  
		 * The command will be executed following the status 
		 * stage of control write. The USB device does not change 
		 * its device address until after the status stage of this
		 * request is completely successfully.
		 */
		TXCNTL = 0x0;	/* Setup Null Packet */
		break;
	case SET_FEATURE:	/* Set Device RWU */
		if( wvalue == DEVICE_REMOTE_WAKEUP )
		{
			SET_HRWUPE;	/* Enable the hub to respond to RWU. */
			TXCNTL = 0x0;	/* Setup Null Packet */
		}
		else
		{
			goto stall_ep0;
		}
		break;
	case CLEAR_FEATURE:	/* Clear Device RWU */
		if ( wvalue == DEVICE_REMOTE_WAKEUP )
		{
			CLR_HRWUPE;	/* Disable the hub to respond to RWU. */
			TXCNTL = 0x0;	/* Setup Null Packet */
		}
		else
		{
			goto stall_ep0;
		}
		break;
	case SET_DESCRIPTOR:
		/* 
		 * Optional, if supported the descriptor may need
		 * updated or a new descriptor may be added.
		 */
	default:		/* Bad set standard device command */
stall_ep0:
		EPCON |= (SET_IN_STALL_MASK | SET_OUT_STALL_MASK);	/* stall EP0 */
		break;
	}
}

/*****************************************************************************
|*
|*  FUNCTION:           get_standard_endpoint_command
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Process a Standard Get Endpoint Command Setup Token.
|*	This can only be a get status or synch frame.
|*
 */
static void get_standard_endpoint_command( void )
{
	unsigned char stall = 0;			 /* assume stall clear */

	switch( bRequest )
	{
	case GET_STATUS:				/* get endpoint status */
		if ( (windex & 0xF) != 1 )		/* not end point 1 stall ? */
		{
			unsigned char save = EPINDEX;	/* save endpoint index */

			/* 
			 * set the endpoint to get status from 
			 */
			EPINDEX = (save & 0x80)|(windex & 0xF);	

			if ( USB_DEVICE_TO_HOST( windex ) )	/* get in stall status */
			{
				if ( GET_TXSTL )
				{
					stall = 1;	/* return stalled transmitter */
				}
			} 
			else				/* get out stall status */
			{
				if ( GET_RXSTL )
				{
					stall = 1;	/* return stalled receiver */
				}
			}

		        EPINDEX  = save;		/* Point the index back to EP0 */
		}
		else if ( GET_EP1STL )			/* Get EP1 Stall */
		{
			stall = 1;			/* return stalled */
		}
	        TXDAT    = stall;
	        TXDAT    = 0x00;
	        TXCNTL   = 0x02;
		break;
	case SYNCH_FRAME:				/* synchronize endpoint */
		/*
		 * Only supported for isochronous transfers, Hubs are
		 * not required to have isochronous endpoints.
		 */
	default:					/* Bad get standard endpoint command */
		EPCON |= (SET_IN_STALL_MASK | SET_OUT_STALL_MASK);	/* stall EP0 */
		break;
	}
}

/*****************************************************************************
|*
|*  FUNCTION:           set_standard_endpoint_command
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Process a Standard Set Endpoint Command Setup Token.
|*	This can only be a Set or Clear Feature Endpoint Stall
|*
 */
static void set_standard_endpoint_command( void )
{
	unsigned char save = EPINDEX;	/* save endpoint index */

	/* 
	 * Get the endpoint index of the stall to clear or set
	 */
	unsigned char index = (windex & 0xF);

	/*
	 * Setup the EPINDEX to point at correct index.
	 */
	EPINDEX = (save & 0x80) | index;

	/*
	 * Check to make sure this is an Endpoint Stall.
	 */
	if ( wvalue == ENDPOINT_STALL )
	{
	        switch( bRequest )
		{
	        case CLEAR_FEATURE: 	/* clear endpoint stall */
			if ( index ) 	/* EP1 clear stall command ? */
			{
				CLR_EP1STL;	/* clear the stall on Hub EP1 */
			}
			else		/* EP0 clear stall command */
			{
				/* 
				 * For EP0 clear both TX & RX stall bits 
				 */
				EPCON &= (CLEAR_IN_STALL_MASK & CLEAR_OUT_STALL_MASK);
			}
			break;
		case SET_FEATURE:	/* set endpoint stall */
			if ( index ) 	/* EP1 set stall command ? */
			{
				SET_EP1STL;    /* set the stall on Hub EP1 */
			}
			else		/* EP0 set stall command */
			{
				goto stall_ep0;	/* just simply stall EP 0 */
			}
			break;
		default:		/* Bad set standard endpoint command */
			goto stall_ep0;
//			break;
		}

		EPINDEX = save;		/* restore endpoint index */
		TXCNTL  = 0x0;		/* Setup Null Packet */
	}
	else
	{
stall_ep0:
		EPINDEX = save;		/* restore endpoint index */
		EPCON |= (SET_IN_STALL_MASK | SET_OUT_STALL_MASK);	/* stall EP0 */
	}
}

/*****************************************************************************
|*
|*  FUNCTION:           get_standard_interface_command
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      This request allows the host to determine the currently selected 
|*	alternative setting of the interface.
|*
 */
static void get_standard_interface_command( void )
{
	if ( bRequest == GET_INTERFACE	&&
	     windex == IDX_INTERFACE_NR0 )
	{
		TXDAT = ALT_INTERFACE_NR0;
	        TXCNTL = 0X01;
	}
	else
	{
		EPCON |= (SET_IN_STALL_MASK | SET_OUT_STALL_MASK);	/* stall EP0 */
	}
}

/*****************************************************************************
|*
|*  FUNCTION:           set_standard_interface_command
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      This request allows the host to select an alternate setting for 
|*	the specified interface. Hubs are only required to support one 
|*	interface.
|*
 */
static void set_standard_interface_command( void )
{
	if ( bRequest == SET_INTERFACE		&&
	     wvalue == ALT_INTERFACE_NR0 	&&
	     windex == IDX_INTERFACE_NR0	)
	{
		TXCNTL = 0x0;		/* Setup Null Packet */
	}
	else
	{
		EPCON |= (SET_IN_STALL_MASK | SET_OUT_STALL_MASK);	/* stall EP0 */
	}
}

/*****************************************************************************
|*
|*  FUNCTION:           get_class_device_command
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*	Process the get hub class device commands,
|*	get descriptor and get status.
|*
 */
static void get_class_device_command( void )
{
	switch( bRequest )
	{
	case GET_DESCRIPTOR:
	        control_buffer( sizeof(HUB_DESCRIPTOR), (unsigned char *)&hub );
		break;
	case GET_STATUS:
	   	TXDAT  = HSTAT & 0x3;		/* Last over-current indicator */
	       	TXDAT  = 0x0;
		TXDAT  = (HSTAT >> 2) & 0x3;	/* Hub over-current indicator  */
        	TXDAT  = 0x0;
	        TXCNTL = 0x4;
		break;
	default:
		EPCON |= (SET_IN_STALL_MASK | SET_OUT_STALL_MASK);	/* stall EP0 */
		break;
	}
}

/*****************************************************************************
|*
|*  FUNCTION:           get_class_other_command
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*	Process the get hub class other commands, get port status and
|*	get bus status.
|*
 */
static void get_class_other_command( void )
{
	unsigned char save = HPINDEX;		/* Save the index */

	HPINDEX = blsbIndex;			/* Select the desired port to get the data from. */

	switch( bRequest )
	{
	case GET_STATUS:			/* Get Port State */
#ifndef INTERNAL_PORT
		if ( blsbIndex == EMBEDDED_PORT_NUMBER )
		{
			/* 
			 * show disconnected
			 */
			TXDAT = 0x0;
			TXDAT = 0x1;	/* only show port powered */
			TXDAT = 0x0;
			TXDAT = 0x0;
			TXCNTL= 0x4;

			/* In reality at this point HPSTAT will show 91 and HPSC 01
			 * PCSTAT = 1 hard-wired, internal function is permanently connected
			 * PCSC = 1 port status change, we need to clear this
			 */
			CLR_PCSC;	/* not port connect status change */
		}
		else
#endif
		{
			/*
			 * Get the first byte of port status
			 */
			if ( GET_PRSTAT )	/* port reset ? */
			{
			      TXDAT = (HPSTAT & 0x7) | 0x10;
			}
			else
			{
			      TXDAT = (HPSTAT & 0x7);
			}
                        /* 
                         * Get second byte of port status
                         */
			if ( GET_PPSTAT )		/* port powered on ? */
			{
				if ( GET_LSSTAT )	/* low-speed device ? */
				{
				        TXDAT = 0x3;	/* powered and low-speed */
				}
				else
				{
				        TXDAT = 0x1;	/* powered and high-speed */
				}
			}
			else			/* powered off */
			{	
				if ( GET_LSSTAT )	/* low-speed device ? */
				{
				        TXDAT = 0x2;	/* powered off and low-speed */
				}
				else
				{	
					TXDAT = 0x0;	/* powered off and high-speed */
				}
			}
                        /* 
                         * Get third byte of port status
                         */
		        TXDAT = (HPSC & 0x1F); /* get the LSB of the wPortChange byte */
		        TXDAT = 0x0;
			TXCNTL= 0x4;
		}
		break;
	case GET_STATE:				/* Get Bus State */ 
	        TXDAT   = (HPSTAT >> 6) & 0x3;	/* Returns bus state, Bit 0 = D-, Bit 1 = D+ */
	        TXCNTL  = 0x1;
		break;
	default:
		EPCON |= (SET_IN_STALL_MASK | SET_OUT_STALL_MASK);	/* stall EP0 */
		break;
	}

        HPINDEX = save;				/* Restore the index */
}

/*****************************************************************************
|*
|*  FUNCTION:           set_class_device_command
|*
|*  AVAILABILITY:	external
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*	Resets a value in the hub status. Only over current can be cleared.
|*
 */
static void set_class_device_command( void )
{
	if ( bRequest == CLEAR_FEATURE )
	{
		if ( bDescriptorIndex == C_HUB_OVER_CURRENT )	/* Check over current */
		{
			CLR_OVISC;	/* clear over-current indicator */
		} 	
		else if ( bDescriptorIndex == C_HUB_LOCAL_POWER ) /* Check local power */
		{
					/* not supported on this device */
		}
		else
		{
			goto stall_ep0;
		}
	}
	else
	{
stall_ep0:
		EPCON |= (SET_IN_STALL_MASK | SET_OUT_STALL_MASK);	/* stall EP0 */
		return;

	}
	TXCNTL = 0x0;			/* setup null packet */
}

/*****************************************************************************
|*
|*  FUNCTION:           set_class_other_command
|*
|*  AVAILABILITY:	external
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*	Process a set port feature or clear port feature
|*
 */
static void set_class_other_command( void )
{
	unsigned char save = HPINDEX;	/* Save the current port */

	HPINDEX = blsbIndex;		/* Select the desired port to get the data from. */

	switch ( bRequest )
	{
	case CLEAR_FEATURE:
		switch(  bDescriptorIndex )
		{
		case PORT_ENABLE:		/* Clear port enable */
			/* also called when device is disabled in
			 * hardware profile from the device manager
			 */
        		HPCON = DISABLE_PORT_CMD;
			break;
		case PORT_SUSPEND:		/* Clear port suspend */
#ifdef INTERNAL_PORT
			if (  bDescriptorIndex == EMBEDDED_PORT_NUMBER )	/* internal port ? */
			{
			        resume();	/* embedded function resume */
			}
#endif
		        HPCON = RESUME_PORT_CMD;
			break;
		case PORT_POWER:		/* Clear port power */
			/*
			 * clear port power masks
		         * 0FFh for 0 = Bit 0 = Port 0, Illegal but we will ignore it.
		         * 0FDh for 1 = Bit 1 = Port 1,
		         * 0FBh for 2 = Bit 2 = Port 2,
		         * 0F7h for 3 = Bit 3 = Port 3
			 * 0EFh for 4 = Bit 4 = Port 4 - Embedded Port
		         * 0DFh for 5 = Bit 5 = Port 5 - Helsinki: 4th downstream port
		         * Use mask to turn off port.
			 */
		        HPPWR &= ~((0x1 << HPINDEX) & 0xFE);
			break;
		case C_PORT_CONNECTION:		/* Clear port connection */
			CLR_PCSC;
			break;
		case C_PORT_ENABLE:		/* Clear port enable */
			CLR_PESC;
			break;
		case C_PORT_SUSPEND:		/* Clear port suspend */
			CLR_PSSC;
			break;
		case C_PORT_RESET:		/* Clear port reset */
	        	CLR_RSTSC;
			break;
		default:			/* Illegal feature selector */
			goto stall_ep0;
//			break;
		}
		break;
	case SET_FEATURE:		/* Set hub other feature */
		switch(  bDescriptorIndex )
		{
		case PORT_ENABLE:	/* Set port enable */
			HPCON = ENABLE_PORT_CMD;
			break;
		case PORT_SUSPEND:	/* Set port suspend */
#ifdef INTERNAL_PORT
			if ( HPINDEX == EMBEDDED_PORT_NUMBER )	/* internal port ? */
			{
				suspend();	/* suspend embedded function */
			}
#endif
			HPCON = SUSPEND_PORT_CMD;
			break;
		case PORT_POWER:	/* Set port power */
			/* Set port power masks
		         * 000h for 0 = Bit 0 = Port 0, Illegal but we will ignore it.
		         * 002h for 1 = Bit 1 = Port 1,
		         * 004h for 2 = Bit 2 = Port 2,
		         * 008h for 3 = Bit 3 = Port 3
			 * 010h for 4 = Bit 4 = Port 4 - Embedded Port
		         * 020h for 5 = Bit 5 = Port 5 - Helsinki: 4th downstream port
		         * Use mask to shut off port.
			 */
		        HPPWR |= (0x1 << HPINDEX) & 0xFE;
			break;	
		case PORT_RESET:	/* Set port reset */
        		HPCON = RESET_ENABLE_PORT_CMD;	/* Do the reset. */
#ifdef INTERNAL_PORT
			if ( HPINDEX == EMBEDDED_PORT_NUMBER )	/* internal port ? */
			{
				/*
				 * This is the internal port, wait for the reset
				 * to complete and then call the software reset
				 * which will configure the hardware as well.
				 *
         			 * Reset is applied at end of frame. The embedded 
         			 * function will actually reset at end of frame and 
         			 * clear all of the USB registers.
				 *
				 */
				while ( !GET_PRSTAT )	/* wait until save to initialize */
					;

				initialize_function();	/* initialize the embedded function. */
			}
#endif
			break;
		default:
			goto stall_ep0;
//			break;
		}
		break;
	default:
stall_ep0:
		HPINDEX = save;	/* Restore the index */
		EPCON |= (SET_IN_STALL_MASK | SET_OUT_STALL_MASK);	/* stall EP0 */
		return;
//		break;
	}

	HPINDEX = save;		/* Restore the index */

       	/*
       	 * single packet only command
	 */
	TXCNTL = 0x0;		/* setup null packet */
}

/*****************************************************************************
|*
|*  FUNCTION:           device_request
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Determine USB Device request from the host on the device's default 
|*	pipe. The request type of the setup packet is processed. For invalid
|*	or reserved commands a stall is generated.
|*
 */
static void device_request( void )
{
	switch( bmRequestType & 0xE3 )
	{
	case	GET_STANDARD_DEVICE_COMMAND:
		get_standard_device_command();
		break;
	case	SET_STANDARD_DEVICE_COMMAND:
		set_standard_device_command();
		break;
	case	GET_STANDARD_ENDPOINT_COMMAND:
		get_standard_endpoint_command();
		break;
	case	SET_STANDARD_ENDPOINT_COMMAND:
		set_standard_endpoint_command();	
		break;
	case	SET_STANDARD_INTERFACE_COMMAND:
		set_standard_interface_command();
		break;
	case	GET_STANDARD_INTERFACE_COMMAND:
		get_standard_interface_command();
		break;

	case	GET_CLASS_DEVICE_COMMAND:
		get_class_device_command();
		break;
	case	GET_CLASS_OTHER_COMMAND:
		get_class_other_command();
		break;
	case	SET_CLASS_DEVICE_COMMAND:
		set_class_device_command();
		break;
	case	SET_CLASS_ENDPOINT_COMMAND:
	case	SET_CLASS_OTHER_COMMAND:
	       	set_class_other_command();
		break;

	case	SET_STANDARD_OTHER_COMMAND:
	case	GET_STANDARD_OTHER_COMMAND:
		/*
	         * Unknown Standard Command -- STALL ENDPOINT
		 */
		/* FALL THROUGH */

	case	SET_CLASS_INTERFACE_COMMAND:
	case	GET_CLASS_INTERFACE_COMMAND:
	case	GET_CLASS_ENDPOINT_COMMAND:
		/*
	         * Unknown Class Command -- STALL ENDPOINT
		 */
		/* FALL THROUGH */

	case	SET_VENDOR_DEVICE_COMMAND:
	case	GET_VENDOR_DEVICE_COMMAND:
	case	SET_VENDOR_INTERFACE_COMMAND:
	case	GET_VENDOR_INTERFACE_COMMAND:
	case	SET_VENDOR_ENDPOINT_COMMAND:
	case	GET_VENDOR_ENDPOINT_COMMAND:
	case	SET_VENDOR_OTHER_COMMAND:		
	case	GET_VENDOR_OTHER_COMMAND:
		/*
	         * Unknown Vendor Command -- STALL ENDPOINT
		 */
		/* FALL THROUGH */

	default:			/* ReservedCommand */
		EPCON |= (SET_IN_STALL_MASK | SET_OUT_STALL_MASK);	/* stall EP0 */
		break;
	}
}

/*****************************************************************************
|*
|*  FUNCTION:           setup_token
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Process a Setup token received on end point 0.
|*
|*	The First step is setting up the Stage tracing variable by examine 
|*	the bmRequestType value. This will tell us if this is a control write
|*	or control read command.
 */
static void setup_token( void )
{
	HRXD0 = 0;		/* Clear the interrupt bit. */

	/*
	 * Now Check for Data still in FIFO.
	 * If this is true then the function
	 * did not see the ACK form the host.
	 */
	 if ( TXFLG & 0xC0 )
	 {
		/* Flush the Transmit FIFOS in case a null packet is still left. */
	 	SET_TXCLR;
		TXCNTL = 0x00;	/* OHCI fix for device Errata */
		SET_TXCLR;	/* OHCI fix for device Errata */
	}

	if ( EPCON & (SET_IN_STALL_MASK | SET_OUT_STALL_MASK) )		/* Stall EP0 ? */
	{
		EPCON  &= ~(SET_IN_STALL_MASK | SET_OUT_STALL_MASK);	/* Clear the stalls */
		TXSTAT |= 0x88;	/* Set data toggle on TX ep0. */
	}

	/*
	 * 8 is length of all setup packets.
	 */
	if ( RXCNTL == 8 )
	{
		/*
		 * Move FIFO to command buffer.
		 */
	       	CLR_EDOVW;

	       	bmRequestType	 = RXDAT;
		bRequest	 = RXDAT;
		bDescriptorIndex = RXDAT;	/* wValue LSB */
		bDescriptortype	 = RXDAT;	/* wValue MSB */
		blsbIndex	 = RXDAT;
		bmsbIndex	 = RXDAT;
		blsbLength	 = RXDAT;
		bmsbLength	 = RXDAT;

	        SET_RXFFRC;	        	/* Update receive FIFO state */

		if ( USB_DEVICE_TO_HOST( bmRequestType ) )	/* Setup Get Command ? */
		{
			stage = READ_STAGE;	/* Advance State Machine to control read stage */
		}
		else				/* Setup Set Command */
		{
			stage = WRITE_STAGE;	/* Advance State Machine to control write state */

			/*
	                 * Now check to see if this is a control write with a data stage.
			 */
			if ( wlength != 0 )
			{
				/* Data stage of control write not on Hub */
				goto rx_setup;
			}
		}

		device_request();
	}

rx_setup:
        CLR_RXSETUP;
}

/*****************************************************************************
|*
|*  FUNCTION:           out_token
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Services all OUT tokens on end point 0.
|*
 */
static void out_token( void )
{
	HRXD0 = 0;			/* Clear the interrupt bit. */

	switch( stage )
	{
	case READ_STAGE:		/* Status stage of control read */
		SET_RXFFRC;		/* Update receive FIFO state */
		stage = SETUP_STAGE;	/* Update the state machine-Expect with a setup packet */
		break;
	case WRITE_STAGE:		/* Data stage of control write not on Hub */
	default:
		EPCON |= (SET_IN_STALL_MASK | SET_OUT_STALL_MASK);	/* stall EP0 */
		break;
	}
}

/*****************************************************************************
|*
|*  FUNCTION:           in_token
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Services all IN tokens on end point 0.
|*
 */
static void in_token( void )
{
	HTXD0 = 0;		/* clear the interrupt bit */

	switch( stage )		/* read state variable */
	{
	case WRITE_STAGE:	/* Status stage of control write */
		if ( bRequest == SET_ADDRESS )
		{
			/* 
			 * Complete standard set address command
			 */
			HADDR = wvalue;
		}
		stage = SETUP_STAGE;	/* Update the state machine-Expect with a setup packet */
		break;
	case READ_STAGE:	/* Data stage of control read */
		control_buffer( 0, (unsigned char *)0 );
		break;
	default:
		EPCON |= (SET_IN_STALL_MASK | SET_OUT_STALL_MASK);	/* stall EP0 */
		break;
	}
}

/*****************************************************************************
|*
|*  FUNCTION:           control_buffer
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*	size	buffer length in bytes
|*	p	pointer to buffer
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Set global buffer pointer, initialize the buffer counter and
|*	copy data from the location pointed to by p.
|*
|*	The number of bytes left to transmit are stored in buffer_size.
|*
|*	If zero bytes are left then a Null packet is loaded even if the 
|*	null/short packet is not needed.
|*
 */
static void control_buffer( unsigned int size, unsigned char * p )
{
	unsigned char		count;		/* Number of bytes in FIFO Always <=16 */
	static unsigned char *  buffer;
	static unsigned int	buffer_size;

	if ( p )				/* Transmit new buffer ? */
	{
		 /*
		  * Only transmit the amount asked for or available
		  */ 
		if ( size > wlength )
		{
	        	size = wlength;		/* amount asked for is smaller then available */
		}
	}
	else					/* continue transmitting current buffer */
	{
		p    = buffer;			/* Get current location */
		size = buffer_size;		/* Get byte left over */
	}

	for ( count = 0; count < EP0_MAX_PACKET_SIZE && size; count++, size-- )
	{
		TXDAT = p[count];		/* Transmit the data */
	}

	buffer      = p+count;			/* Update buffer pointer */
	buffer_size = size;			/* Update buffer size left */

        TXCNTL = count;				/* Write count into TXCNT register */
	SET_TXOE;				/* Enable data transmit */
}

/*****************************************************************************
|*
|*  FUNCTION:           SOF_Hub_isr
|*
|*  AVAILABILITY:	extern
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*	The HX shares this ISR with the hub.
|*
|*      Service the Start Of Frame Interrupt
|*
|*	This routine simply displays the upper three bytes of the
|*	SOF in the lower three bits of the LED's on P1.
|*	It does not affect the other LED's. Very useful
|*	in determining when the function is receiving SOFs
|*
|*	Service the HUB Interrupt
|*
|*	This routine scans the interrupt pending bits in order
|*	and branches to any routines that have pending interrupts.
|*	After finishing the routine jumps back to the top of
|*	the scanning loop to preserve interrupt priority.
|*
|*	NOTE: The order of precedence is set by the order in which the
|*	      user places the checks.
|*
 */
void _interrupt( 8 ) _using( 0 ) SOF_Hub_isr( void )
{
	unsigned char save;
#ifdef INTERNAL_PORT
	if ( GET_ASOF )	/* If this ASOF bit not set, the ISR could be a HUB.  Go Check. */
	{
		CLR_ASOF;

		P1_0 = GET_TS8;
		P1_1 = GET_TS9;
		P1_2 = GET_TS10;

		/* 
	         * Check to see if there is a HUB pending ISR
	         * By checking this last, we give the SOF interrupt
	         * higher priority.
		 */
		if ( !( HRXD0 || HTXD0 ) )
			return;
	}
#endif

	save = EPINDEX;	/* save index */

	/*
	 * Process Hub End points
	 */
	while ( HRXD0 || HTXD0 )
	{    
		/* 
		 * If this is an EP0 command then we need to process TX & RX
		 * ISRs in the correct order if, for whatever reason they 
		 * occurred back to back. This happens in OHCI systems.
		 */
	        EPINDEX = 0x80;

		/* Now check to see what stage the TX/RX is in.
                 * If it's in the control read stage then check TX first.
                 * If it's in the setup or control write stage then check the RX first.
		 */
		if ( stage == READ_STAGE )
		{
			/*
			 * Process the TX FIFO first
                         */
			if ( HTXD0 )
			{
			        in_token();	/* 8293X has sent a packet to HC */
			}
			else /* if ( HRXD0 ) */
			{
				if ( GET_RXSETUP )		/* Is this a setup packet ? */
				{
					setup_token();
				}
				else
				{
				       	out_token();	/* 8293X has received a packet from HC */
				}
			}
		 }
		 else
		 {
			/*
			 * Process the RX FIFO first
			 */
			if ( HRXD0 )
			{
				if ( GET_RXSETUP )		/* Is this a setup packet ? */
				{
					setup_token();
				}
				else
				{
				       	out_token();	/* 8293X has received a packet from HC */
				}
			}
			else /* if ( HTXD0 ) */
			{
			        in_token(); 	/* 8293X has sent a packet to HC */
			}
		}
		/* Look at all Interrupt flags again */
	}

	EPINDEX = save;	/* restore index */
}

#ifndef INTERNAL_PORT
/*****************************************************************************
|*
|*  FUNCTION:           suspend_resume_isr
|*
|*  AVAILABILITY:	external
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Service the suspend and resume interrupt service routine.
|*
 */
void _interrupt( 10 ) _using( 0 ) suspend_resume_isr( void )
{
	SET_PD;		/* power down */


	/*
	 * Let device know that a resume signal was detected.
	 */
	CLR_GRSM;	/* Clear resume flag */
	CLR_GSUS;	/* Clear suspend flag */
}
#endif
