/*****************************************************************************
|*
|*  VERSION CONTROL:    @(#)loopback.c	1.1 7/3/98
|*
|*  IN PACKAGE:         AX/HX93X Firmware
|*		 
|*  COPYRIGHT:          Copyright (c) 1998 Tasking Software B.V.
|*
|*  DESCRIPTION:
|*
|*	Universal serial bus AX/HX93X embedded function loop back firmware
|*
 ****************************************************************************/

#include "eval.h"
#include "usb.h"
#include "loopback.h"

#if !defined( _C251 ) && !defined( _CC51 )
#define _using( X )
#endif


/*
 *	function_isr loop back subroutines
 */
#if ( EPPS > EPP_OFF )
static unsigned char * fifo_RX_TX		( unsigned char *, unsigned char *,
						  unsigned char, unsigned int,
						  unsigned char );
static unsigned int ep_loop_back_TX		( unsigned char *, unsigned char **, 
				     	          unsigned int, unsigned int, 
						  unsigned char, 
						  unsigned int );
static unsigned int ep_loop_back_RX		( unsigned char *, unsigned char **, 
				     	          unsigned int, unsigned int, 
						  unsigned char, 
						  unsigned int );
#endif


/*****************************************************************************
|*
|*  FUNCTION:           LOOP
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*	n	endpoint number
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*	Loop back interrupt service macro. This macro calls the
|*	endpoint loop back isr functions for the specified endpoint 
|*	number. And allocates the storage for the endpoint count,
|*	the circular endpoint loop back buffer, the endpoint buffer
|*	read and write pointers.
|*
 */
#define	LOOPBACK( n ) 											\
{													\
	static unsigned int	ep ## n ## _count = 0;							\
	static unsigned char	ep ## n ## _fifo[EP  ## n ## _LOOP_BACK_BUFFER_SIZE];			\
	static unsigned char *	ep ## n ## _write = ep ## n ## _fifo;					\
	static unsigned char *	ep ## n ## _read  = ep ## n ## _fifo;					\
													\
	if ( FTXD ## n )										\
	{												\
		EPINDEX = n;										\
		ep ## n ## _count = ep_loop_back_TX( ep ## n ## _fifo, 					\
						    &ep ## n ## _read, 					\
						     ep ## n ## _count,					\
					     	     EP ## n ## _LOOP_BACK_BUFFER_SIZE, 		\
					     	     EP ## n ## _LOOP_BACK_FIFO_SIZE, 			\
						     ~(1 << (2 * n)) );					\
		/* 											\
		 * If data is pending in the FIFOs go get it. 						\
		 */											\
		if ( GET_RXFIF0 || GET_RXFIF1 )									\
		{											\
			/* 8293X has received a packet from HC */					\
			ep ## n ## _count = ep_loop_back_RX( ep ## n ## _fifo, 				\
							    &ep ## n ## _write, 			\
							     ep ## n ## _count, 			\
						     	     EP ## n ## _LOOP_BACK_BUFFER_SIZE,		\
						     	     EP ## n ## _LOOP_BACK_FIFO_SIZE, 0 );	\
		}											\
	}												\
	if ( GET_FRXD ## n )										\
	{      												\
		EPINDEX = n;										\
		/* 											\
		 * 8293X has received a packet from HC							\
		 */											\
		ep ## n ## _count = ep_loop_back_RX( ep ## n ## _fifo, 					\
						    &ep ## n ## _write, 				\
						     ep ## n ## _count, 				\
					     	     EP ## n ## _LOOP_BACK_BUFFER_SIZE,			\
					     	     EP ## n ## _LOOP_BACK_FIFO_SIZE, 			\
		        			     ~(1 << (1+(2 * n))) );				\
		ep ## n ## _count = ep_loop_back_TX( ep ## n ## _fifo, 					\
						    &ep ## n ## _read, 					\
						     ep ## n ## _count,					\
					     	     EP ## n ## _LOOP_BACK_BUFFER_SIZE, 		\
					     	     EP ## n ## _LOOP_BACK_FIFO_SIZE, 0 );		\
	}												\
}

/*****************************************************************************
|*
|*  FUNCTION:           loopback
|*
|*  AVAILABILITY:	external
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*	The loop back dispatch function
|*
 */
 void loopback( void )
 {
#if ( EPPS >= 1 ) && ( EPP_OFF < 1 )
		LOOPBACK( 1 );	/* loop back via endpoint 1 */
#endif
#if ( EPPS >= 2 ) && ( EPP_OFF < 2 )
		LOOPBACK( 2 );	/* loop back via endpoint 2 */
#endif
#if ( EPPS >= 3 ) && ( EPP_OFF < 3 )
		LOOPBACK( 3 );	/* loop back via endpoint 3 */
#endif
#if defined( _REG930AX_SFR ) && ( EPPS >= 4 ) && ( EPP_OFF < 4 )
		LOOPBACK( 4 );	/* loop back via endpoint 4 */
#endif
#if defined( _REG930AX_SFR ) && ( EPPS >= 5 ) && ( EPP_OFF < 5 )
		LOOPBACK( 5 );	/* loop back via endpoint 5 */
#endif
}

#if ( EPPS > EPP_OFF  )
/*****************************************************************************
|*
|*  FUNCTION:           fifo_RX_TX
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*	fifo	pointer to fifo
|*	p	pointer to current location in fifo
|*	count	number of byte to receive or transmit
|*	size	size of the fifo
|*	receive	True if data received in fifo
|*		False if data transmit from fifo
|*
|*  RETURN VALUE:
|*
|*	Pointer to current fifo location
|*
|*  DESCRIPTION:
|*
|*      Receive or Transmit data from fifo
|*
 */
static unsigned char * fifo_RX_TX( unsigned char * fifo, unsigned char * p, 
				   unsigned char count, unsigned int size, 
				   unsigned char receive )
{
	unsigned char * end = fifo + size;

	for ( count; count != 0; count-- )
	{ 
		if ( receive )	/* receive data */
		{
			*p = RXDAT;
		}		
		else		/* transmit data */
		{
			TXDAT = *p;
		}

		if ( p == end )
		{
			/* If this is equal then we need to move the write
			 * pointer to the beginning of the fifo
			 */
			p = fifo;
		}
		else
		{
			p++;
		}
	}

	return p;
}

/*****************************************************************************
|*
|*  FUNCTION:           ep_loop_back_RX
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*	fifo	pointer to circular fifo
|*	p	pointer to pointer to current location in circular fifo
|*	count	number of byte to receive
|*	size	circular fifo size
|*	maxsz	maximum packet size
|*	mask	0 for Pre-receive operations
|*		x interrupt mask for Post-receive operations on pipe (1-n)
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*	End point receiver loop back function
 */
static unsigned int ep_loop_back_RX( unsigned char * fifo, unsigned char ** p, 
		                     unsigned int count, unsigned int size, 
			             unsigned char maxsz, unsigned int mask )
{
	if ( mask )			/* Post-receive operation */
	{  
 		FIFLG &= mask;		/* Clear interrupt pending flag */
#if defined( _REG930AX_SFR ) && ( EPPS > 3 )
 		FIFLG1 &= (mask >> 8);
#endif
		if ( GET_RXERR )		/* Receiver error ? */
		{  
			if ( GET_RXOVF )	/* Receiver overflow ? */
			{  
				CLR_RXOVF; /* Clear sticky bit */
				SET_RXCLR; /* Flush Receive FIFO */
				return 0;
			}
			/* 
			 * Else if failed CRC or Bit-stuffing
			 * then Reverse Receive FIFO to Current 
			 * Packet Entry
			 */
		}
	}

	/* 
	 *	Post/Pre-receive operation
	 */
	while ( ( GET_RXFIF0 || GET_RXFIF1 )	   &&
		( count < (size - maxsz) ) )
	{
		unsigned char rxcntl = RXCNTL;	/* Get number of bytes in FIFO */
		/*
		 * Read data Packet(s) from Receive FIFO through RXDAT
		 */
		*p = fifo_RX_TX( fifo, *p, rxcntl, size-1, 1 );

	       	count += rxcntl;

		if ( GET_RXURF ) 			/* Receiver underrun */
		{      
			CLR_RXURF;		/* Clear sticky bit */
			return 0;
		}

	        SET_RXFFRC;			/* Release the FIFO */
	}
	return count;
}

/*****************************************************************************
|*
|*  FUNCTION:           ep_loop_back_TX
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*	fifo	pointer to circular fifo
|*	p	pointer to pointer to current location in circular fifo
|*	count	number of byte to transmit
|*	size	circular fifo size
|*	maxsz	maximum packet size
|*	mask	0 for Pre-transmit operations
|*		x interrupt mask for Post-transmit operations on pipe (1-n)
|*
|*  RETURN VALUE:
|*
|*	number of bytes written to transmit FIFO
|*
|*  DESCRIPTION:
|*
|*	End point transmitter loop back function
 */
static unsigned int ep_loop_back_TX( unsigned char * fifo, unsigned char ** p, 
		                     unsigned int count, unsigned int size, 
			             unsigned char maxsz, unsigned int mask )
{
	if ( mask )			/* Post-transmit operation */
	{
 		FIFLG &= mask;		/* Clear interrupt pending flag */
#if defined( _REG930AX_SFR ) && ( EPPS > 3 )
 		FIFLG1 &= (mask >> 8);
#endif

		if ( GET_TXERR )
		{
			if ( GET_TXURF )
			{
				CLR_TXURF;	/* Clear sticky bit */
				SET_TXCLR;	/* Flush Transmit FIFO */
			}
			/* 
			 * Else if failed CRC, Bit-stuffing,
			 * or Timeout from Host	then Reverse 
			 * Transmit FIFO to Current Packet Entry
			 */
			return 0; 
		}
	}

	/* 
	 *	Pre/Post-transmit operation
	 */
	while ( !( GET_TXFIF0 || GET_TXFIF1 ) && 
		 ( count != 0 ) 	)
	{
		unsigned int cnt;

		if ( count > maxsz )
		{
			cnt = maxsz;
		}
		else
		{
			cnt = count;
		}

		/*
		 * Transfer Packet to Transmit FIFO Through TXDAT
		 */
		*p = fifo_RX_TX( fifo, *p, cnt, size-1, 0 );

		count -= cnt;

		if ( GET_TXOVF )		/* Transmitter overflow ? */
		{
			CLR_TXOVF;	/* Clear sticky bit */
			SET_TXCLR;	/* Flush Transmit FIFO */
			return 0;
		}
		TXCNTL = cnt;		/* Release the FIFO */
	}

	return count;
}
#endif
