/*****************************************************************************
|*
|*  VERSION CONTROL:    @(#)loopback.h	1.1 7/3/98
|*
|*  IN PACKAGE:		AX/HX93X Firmware
|*
|*  COPYRIGHT:          Copyright (c) 1998 Tasking Software B.V.
|*
|*  DESCRIPTION:
|*
|*	Universal serial bus AX/HX93X embedded loopback function definitions
|*
 ****************************************************************************/

#ifndef _LOOPBACK_H
#define _LOOPBACK_H

void loopback			( void );

#ifndef LOOP_BACK_BUFFER_SIZE
#define	LOOP_BACK_BUFFER_SIZE		0x0400
#endif

#ifdef _SINGLE_PACKET_MODE
#define _DIV 1 /* single packet mode */
#else 
#define _DIV 2 /* dual packet mode */
#endif

/*
 * Maximum Transmit/receive FIFO sizes for dual packet mode
 *
 * When operating in dual packet mode, the maximum packet size should be,
 * at a maximum, half the FIFO size to ensure both wil simultaneously fit
 * in the FIFO.
 */
#if defined( _REG931HX_SFR ) || defined( _REG931AX_SFR )
#define EP1_LOOP_BACK_FIFO_SIZE		(16 / _DIV) /* endpoint 1 loop back fifo size for 8x931 */
#else
#define EP1_LOOP_BACK_FIFO_SIZE		(256 / _DIV)  /* endpoint 1 loop back fifo size for 8x930 */
#endif

#if defined( _REG931HX_SFR ) || defined( _REG931AX_SFR )
#define EP2_LOOP_BACK_FIFO_SIZE		(8 / _DIV) /* endpoint 2 loop back fifo size for 8x931 */
#else 
#if EPPS > 3
#define EP2_LOOP_BACK_FIFO_SIZE		(32 / _DIV) /* endpoint 2 loop back fifo size for 8x930 with 6 epp */
#else 
#define EP2_LOOP_BACK_FIFO_SIZE		(16 / _DIV) /* endpoint 2 loop back fifo size for 8x930 with less then 6 epp */
#endif
#endif

#if defined( _REG930HX_SFR ) || defined( _REG930AX_SFR )
#if EPPS > 3 
#define EP3_LOOP_BACK_FIFO_SIZE		(32 / _DIV) /* endpoint 3 loop back fifo size for 8x930 with 6 epp */
#else
#define EP3_LOOP_BACK_FIFO_SIZE		(16 / _DIV) /* endpoint 3 loop back fifo size for 8x930 with less then 6 epp */
#endif
#else 
#define EP3_LOOP_BACK_FIFO_SIZE		0 /* endpoint 3 loop back fifo size for 8x931 not possible */
#endif

#if defined( _REG930AX_SFR )
#if EPPS > 3 
#define EP4_LOOP_BACK_FIFO_SIZE		(32 / _DIV) /* endpoint 4 loop back fifo size for 8x930 with less then 6 epp */
#endif
#else 
#define EP4_LOOP_BACK_FIFO_SIZE		0 /* endpoint 4 loop back fifo size for 8x931 8x930HX not possible */
#endif

#if defined( _REG930AX_SFR ) 
#if EPPS > 3 
#define EP5_LOOP_BACK_FIFO_SIZE		(16 / _DIV) /* endpoint 5 loop back fifo size for 8x930 with less then 6 epp */
#endif
#else 
#define EP5_LOOP_BACK_FIFO_SIZE		0 /* endpoint 5 loop back fifo size for 8x931 8x930HX not possible */
#endif


#define	EP1_LOOP_BACK_BUFFER_SIZE	LOOP_BACK_BUFFER_SIZE + EP1_LOOP_BACK_FIFO_SIZE + 1  /* 256 + 64 + 1 */
#define	EP2_LOOP_BACK_BUFFER_SIZE	LOOP_BACK_BUFFER_SIZE + EP2_LOOP_BACK_FIFO_SIZE + 1  /* 256 + 8  + 1 */
#define	EP3_LOOP_BACK_BUFFER_SIZE	LOOP_BACK_BUFFER_SIZE + EP3_LOOP_BACK_FIFO_SIZE + 1  /* 256 + 8  + 1 */
#define	EP4_LOOP_BACK_BUFFER_SIZE	LOOP_BACK_BUFFER_SIZE + EP4_LOOP_BACK_FIFO_SIZE + 1  /* 256 + 8  + 1 */
#define	EP5_LOOP_BACK_BUFFER_SIZE	LOOP_BACK_BUFFER_SIZE + EP5_LOOP_BACK_FIFO_SIZE + 1  /* 256 + 8  + 1 */

/*
 *	Endpoint offset EPP_OFF must be set to 1 when embedded function
 *	is used in combination with hub firmware code. Endpoint 1 is
 *	used to inform the host of a hub or port status change.
 */
#ifndef EPP_OFF
#define EPP_OFF	0
#endif

/*
 *	Set and check number of endpoints used
 */
#ifndef EPPS
#define EPPS	3			/* default 3 endpoints supported */
#else
#if defined( _REG930AX_SFR ) 
#if ( EPPS > 5 )			/* AX930 supports up to 5 end points + EP0 */
illegal number of endpoints
#endif
#else
#if defined( _REG931AX_SFR ) || defined(_REG931HX_SFR)
#if ( EPPS > 2 )			/* 931 supports up to 2 end points + EP0 */
illegal number of endpoints
#endif
#else
#if ( EPPS > 3 )			/* HX930 supports up to 3 end points + EP0 */
illegal number of endpoints
#endif
#endif
#endif
#endif

/*
 * Prevent negative number of endpoints
 */
#if (( EPPS - EPP_OFF ) < 0 )
#define EPPS EPP_OFF
#endif

#endif /* _LOOPBACK_H */
